
configparams updateparameter;

typedef void (*updatefunction_ptr_T)(char*);
updatefunction_ptr_T updatefunction_ptr;

void copy_command(char * command_in, char * data_in){
	// Copy the contents of data_in into command_in
	ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
		if (strnlen (data_in, DATABUFFERLEN)){
			memcpy(command_in, data_in, strnlen (data_in, DATABUFFERLEN-1));
			command_in[(strnlen (data_in, DATABUFFERLEN-1))]=0x0;//Zero terminate the string
		}
		else {
			command_in[0]=0x0;
		}
		// Now clear data_in, the UART can reuse it now
		memset(data_in, 0, strnlen (data_in, DATABUFFERLEN));
	}
}



void process_serial(){
	static unsigned char data_count = 0;
	static unsigned char data_in[DATABUFFERLEN];

	while (Serial.available()){
		// Add char to input buffer
		if(data_count<DATABUFFERLEN-1){
			data_in[data_count] =  Serial.read();
			// 		Serial.print(data_in[data_count],HEX);
			// 		Serial.print(F(" "));
		}

		else{
			data_in[DATABUFFERLEN-1]=0; //make sure the string is zero terminated
			data_count = 0;
			Serial.println(F("Entry to long, try again!"));
		}
		// "^" --> escape
		// "[" --> Start of escape seqence
		// "PgUP" --> PageUP
		// "PgUP" --> PageDown
		// "PgUP" --> PageUP
		// "PgUP" --> PageDown
		
		//	(data_in[data_count] == 0x35) || \
		//	(data_in[data_count] == 0x36) || \
		

		if((data_in[data_count] == 0xd) && (serialstate & (1<<SERIALDATAENTRY))){ //data entry active and carriage return
			//Serial.print(F("CR Pressed"));
			if (data_count == 0){
				Serial.println("");
				data_in[data_count]=0x0; //Zero terminate empty String
				process_command(data_in);
			}
			else{
				//Serial.println("Count: ");
				//Serial.println(data_count);
				data_in[data_count]=0x0; //Zero terminate String (and overwrite CR)
				process_command(data_in);
				data_count = 0;
			}
		}
		else if((data_in[data_count] == 0x1b) && (serialstate & (1<<SERIALDATAENTRY))){ //Escape was pressed during data entry
			data_in[0]=0x1b;	//Ecape Char is first all other data is rellevant
			data_in[1]=0x0;		//Zero terminate empty String
			process_command(data_in);
			data_count = 0;
		}
		else if (serialstate & (1<<SERIALDATAENTRY)){ //normal data entry active
			
			if ((data_in[data_count] >= 0x30 && data_in[data_count] <= 0x39)||(data_in[data_count] >= 0x61 && data_in[data_count] <= 0x66)){ //Sanity Check, only hex entry is allowed
				Serial.print((char)data_in[data_count]);	//local Loopback
				data_count++;								//increment data count index
			}
			else if(data_in[data_count] == 0x7f){
				Serial.print((char)data_in[data_count]);	//local Loopback
				data_in[data_count] = 0x0;					//Clear Del Char from buffer
				if(data_in){
					data_count--;
				}
			}
			else{
				data_in[data_count]=0x0;
			}
		}
		// 0x1b --> "^" --> escape
		else if (data_in[data_count] == 0x1b){ //|| \ (data_in[data_count] == 0x5b)){
			//Data Should be entered
			serialstate |= (1<<SERIALESCAPE);
			data_count++;	//increment data count index
		}
		else if ( data_count == 2 && (serialstate & (1<<SERIALESCAPE)) && data_in[data_count] >= 0x41 && data_in[data_count] <= 0x44){ //3rd character in ESS secquence, Escape sequence is active,  key is arrow (not tilde terminated)
			process_command(data_in);
			serialstate &= ~(1<<SERIALESCAPE);
			data_count = 0;
		}
		else if ((serialstate & (1<<SERIALESCAPE)) && (data_in[data_count] != 0x7E) && data_count<5){ // Escape sequence is active, next key is not the escape terminator (tilde ~), next key is not arrow (not tilde terminated), and some lengh check for safty
			data_count++;	//increment data count index
		}
		else{
			process_command(data_in);
			serialstate &= ~(1<<SERIALESCAPE);
			data_count = 0;
		}

	}
}

void process_command(char * data_in){
	static unsigned char cmd_in[DATABUFFERLEN];
	copy_command(cmd_in, data_in);					//First free up the serial buffer 
	if (serialstate & (1<<SERIALDATAENTRY)){
		if(strlen(cmd_in)){
			//serialstate |= (1<<SERIALDATAREADY);
			//Serial.print(F("Data entered: "));
			//Serial.println((char *)cmd_in);
			updatefunction_ptr(cmd_in);
			// 			for (uint8_t i = 0; i < DATABUFFERLEN; i++){
			// 				Serial.print(cmd_in[i],HEX);
			// 				Serial.print(F(" "));
			// 			}
			//			Serial.println(F(""));
		}
		else{
			updatefunction_ptr(NULL);
			//Serial.print(F("No Data entered!"));
		}
		//Serial.println(F(""));
		serialstate &= ~(1<<SERIALDATAENTRY);
	}
	else if (!(((serialstate & (1<<SERIALDATAENTRY))) || serialstate & (1<<SERIALCONFIG))){		//This Case selection will be executed when on main screen
		uint8_t error;
		uint8_t updatecount = 0;
		switch (cmd_in[0]){
			case '1':
				updateparameter = P_SETBANK;
				updatefunction_ptr = &updateconfighandler;
				serialstate |= (1<<SERIALDATAENTRY);
				Serial.print(F("Enter 1 or 2 for local Backup EEPROM Bank: "));
			break;
			case '2':
				if (eeprombank >= 1 && eeprombank <= 2){
					for (uint8_t i = 0 ; ; i++) {
// 						Serial.print(i);
// 						Serial.print(F(":"));
						printHex(EEPROM.readByte(i+((eeprombank-1)*256)),2); 
						Serial.print(" ");
						if ((i + 1) % 16 == 0) Serial.println();
						if (i == 255) break;
					}
					Serial.println();
				}
				else{
					printselecteeprommsg();
				}
			break;
			case '3':
			if (eeprombank >= 1 && eeprombank <= 2){
				Wire.beginTransmission(I2C_ADDR);
				error = Wire.endTransmission();
				if (error == 0){
					for (uint8_t i = 0 ; ; i++) {
						if (EEPROM.updateByte(i+((eeprombank-1)*256),readByte(I2C_ADDR,i))){
							updatecount++;
						}
						if (i == 255) break;
					}
					Serial.print(updatecount);
					Serial.println(F(" Bytes Written to local EEPROM!"));
				}
				else{
					Serial.print(F("Error Reading from I2C EEPROM @ 0x"));
					Serial.println(I2C_ADDR, HEX);
				}
			}
			else{
				printselecteeprommsg();
			}
			break;
			case '4':
			if (eeprombank >= 1 && eeprombank <= 2){
				updateparameter = P_CONFIRMRESTORE;
				updatefunction_ptr = &updateconfighandler;
				serialstate |= (1<<SERIALDATAENTRY);
				Serial.print(F("Enter \"4321\" to confirm Restoring local EEPROMBANK "));
				Serial.print(eeprombank);
				Serial.print(F(" to meters EEPROM: "));
			}
			else{
				printselecteeprommsg();
			}
			case '5':
			if (eeprombank >= 1 && eeprombank <= 2){
				Wire.beginTransmission(I2C_ADDR);
				error = Wire.endTransmission();
				if (error == 0){
					for (uint8_t i = 0 ; ; i++) {
						uint8_t localeebyte = EEPROM.readByte(i+((eeprombank-1)*256));
						uint8_t i2ceebyte = readByte(I2C_ADDR,i);
						if (localeebyte != i2ceebyte){
							Serial.print(F("Data @ Address 0x"));
							Serial.print(i, HEX);
							Serial.print(F(" Local: 0x"));
							printHex(localeebyte, 2);
							Serial.print(F(" Meter: 0x"));
							printHex(i2ceebyte, 2);
							Serial.println();
							updatecount++;
						}
						if (i == 255) break;
					}
					Serial.print(updatecount);
					Serial.println(F(" Bytes Different between local and meters EEPROM!"));
				}
				else{
					Serial.print(F("Error Reading from I2C EEPROM @ 0x"));
					Serial.println(I2C_ADDR, HEX);
				}
			}
			else{
				printselecteeprommsg();
			}
			break;
			case 'S':
			case 's':
				scanI2C();
			break;
			case 'd':
			case 'D':
				dumpEEPROM();
			break;
			case '?':
			printHelp();
			break;
			case 'r':
			case 'R':
				if ((serialstate & (1<<SERIALNEWADDRESSENTRYVALID))){
					Wire.beginTransmission(I2C_ADDR);
					error = Wire.endTransmission();
					if (error == 0){
							uint8_t data = readByte(I2C_ADDR, neweepromaddress);
							//Serial.print(b, HEX); Serial.print(" ");
							Serial.print(F("Data @ 0x"));
							Serial.print(neweepromaddress, HEX);
							Serial.print(F(": "));
							Serial.println(data, HEX);
					}
					else {
						Serial.print(F("Error Reading from I2C EEPROM @ 0x"));
						Serial.println(I2C_ADDR, HEX);
					}
				}
				else{
					Serial.println(F("Enter Read Address first"));
				}
			break;
			case 'A':
			case 'a':
				updateparameter = P_ADDRESS;
				updatefunction_ptr = &updateconfighandler;
				serialstate |= (1<<SERIALDATAENTRY);
				Serial.print(F("Enter Data Read/Write Address (HEX): "));
			break;
			case 'b':
			case 'B':
				updateparameter = P_DATA;
				updatefunction_ptr = &updateconfighandler;
				serialstate |= (1<<SERIALDATAENTRY);
				Serial.print(F("Enter Data Byte (HEX): "));
			break;
			case 'w':
			case 'W':
				updateparameter = P_CONFIRMWRITE;
				if ((serialstate & ((1<<SERIALNEWDATAENTRYVALID)|(1<<SERIALNEWADDRESSENTRYVALID)))==(1<<SERIALNEWDATAENTRYVALID)|(1<<SERIALNEWADDRESSENTRYVALID)){
					updatefunction_ptr = &updateconfighandler;
					serialstate |= (1<<SERIALDATAENTRY);
					Serial.print(F("Enter \"1234\" to confirm writing: "));
				}
				else{
					Serial.println(F("Enter Address and New Data First!"));
				}
			break;
			case 0x1b: //Escape command "^"
			if (cmd_in[1] == 0x5b) //Escape Sequence start "["
			{
				switch (cmd_in[2]){
					case 0x31:
					if (cmd_in[4] == 0x7E){ //Long Escape Sequence (i.e. F1-F12 Keys - F1 --> 1B 5B 31 31 7E, F12 -> 1B 5B 32 34 7E)
						switch (cmd_in[3]){
							//case 0x31: //F1
							//break;
							
							default:
							#ifdef SERIALBUTTONDEBUG
							Serial.print(F("Escape Command Pressed: "));
							Serial.print(cmd_in[2], HEX);
							Serial.print(F(" "));
							Serial.println(cmd_in[3], HEX);
							#endif
							break;
						}
					}
					break;
					case 0x32:
					if (cmd_in[4] == 0x7E){ //Long Escape Sequence (i.e. F1-F12 Keys - F1 --> 1B 5B 31 31 7E, F12 -> 1B 5B 32 34 7E)
						switch (cmd_in[3]){
							//case 0x34: //F12
							//break;
							
							default:
							#ifdef SERIALBUTTONDEBUG
							Serial.print(F("Escape Command Pressed: "));
							Serial.print(cmd_in[2], HEX);
							Serial.print(F(" "));
							Serial.println(cmd_in[3], HEX);
							#endif
							break;
						}
					}
					#ifdef SERIALBUTTONDEBUG
					else {
						Serial.print(F("Escape Command Pressed: "));
						Serial.println(cmd_in[2], HEX);
					}
					#endif
					
					break;
// 					case 0x35: //PageUP
// 					break;
// 					case 0x36: //PageDown
// 					break;
// 					case 0x41: //ArrowUp
// 					break;
// 					case 0x42: //ArrowDown
// 					break;
					default:
					#ifdef SERIALBUTTONDEBUG
					Serial.print(F("Escape Command Pressed: "));
					Serial.println(cmd_in[2], HEX);
					#endif
					break;
				}
			}
			break;
			
			default:
			#ifdef SERIALBUTTONDEBUG
			Serial.print(F("Single Command Pressed: "));
			Serial.println(cmd_in[0], HEX);
			#endif
			break;
		}
	}
}

void printHelp(){
	VT100helper_MoveTo(1,1);
	VT100helper_ClearCurserDown();
	Serial.println(F("UT21x Dumper and config Tool partly written by RDOB"));
	Serial.println(F("Function Description:"));
	Serial.println(F("[1]: Select local EEPROM Backup Bank (1 | 2)"));
	Serial.println(F("[2]: Dump Local EEPROM Content [3]: Backup Meters EEPROM to local Bank"));
	Serial.println(F("[4]: Restore Local EEPROM Bank to Meter"));
	Serial.println(F("[5]: Show Diff between Meter and local backup"));
	Serial.println(F("[a]: Enter Meters EEprom address to read/write; [r]: Read Byte from Meter"));
	Serial.println(F("[b]: Enter Byte Data (to write to Meter); [w]: Write Byte to Meters EEPROM"));
	Serial.println(F("[d]: Dump Meters EEprom to screen [?]: Print this Help!"));
	Serial.println(F("[s]: Scan I2C Bus for available Devices"));
	Serial.println("");
}

void printselecteeprommsg(){
	Serial.print(F("Select EEPROM Bank First!"));
}


void printenterdatamsg(){
	Serial.print(F("Enter New Value: "));
	}

void updateconfighandler(char * newconfdata){
	if (!newconfdata){
		//VT100helper_MoveTo(16,1);
		//VT100helper_ClearCurserDown();
		Serial.println(F("\n\rNo Data Entered"));
	}
	else if (newconfdata[0]==0x1b){
		VT100helper_MoveTo(16,1);
		VT100helper_ClearCurserDown();
		Serial.println(F("\n\rData Entry Canceled!"));
	}
	else if (isxdigit(*newconfdata)) {
		uint16_t confirmval = 0;
		uint8_t updatecount = 0;
		switch(updateparameter){
			case P_DATA:
			neweepromdata = strtol(newconfdata, &newconfdata, 16);
			Serial.print(F("\n\rNew EEPROM Data: 0x"));
			Serial.println(neweepromdata, HEX);
			serialstate |= (1<<SERIALNEWDATAENTRYVALID);
			break;
			case P_ADDRESS:
			neweepromaddress = strtol(newconfdata, &newconfdata, 16);
			Serial.print(F("\n\rNew EEPROM Read/Write Address: 0x"));
			Serial.println(neweepromaddress, HEX);
			serialstate |= (1<<SERIALNEWADDRESSENTRYVALID);
			break;
			case P_CONFIRMWRITE:
			confirmval = strtol(newconfdata, &newconfdata, 10);
			if (confirmval == 1234){
				Wire.beginTransmission(I2C_ADDR);
				uint8_t error = Wire.endTransmission();
				if (error == 0){
				writeByte(I2C_ADDR, neweepromaddress, neweepromdata);
				Serial.println(F("\n\rData Written!"));
				}
				else{
					Serial.print(F("Error Writing to I2C EEPROM @ 0x"));
					Serial.println(I2C_ADDR, HEX);
				}
			}
			else{
				Serial.println(F("\n\rWriting canceled!"));
			}
			break;
			case P_SETBANK:
			confirmval = strtol(newconfdata, &newconfdata, 10);
			if (confirmval >= 1 && confirmval <= 2){
				eeprombank = confirmval;
				Serial.print(F("\n\rSelected local EEPROM Bank: "));
				Serial.println(eeprombank);
			}
			else{
				Serial.println(F("\n\rInvalid Bank # entered!\r\n"));
			}
			break;
			case P_CONFIRMRESTORE:
			confirmval = strtol(newconfdata, &newconfdata, 10);
			if (confirmval == 4321){
				Wire.beginTransmission(I2C_ADDR);
				uint8_t error = Wire.endTransmission();
				if (error == 0){
					for (uint8_t i = 0 ; ; i++) {
						digitalWrite(LED_BUILTIN, HIGH);
						uint8_t metereebyte = readByte(I2C_ADDR,i);
						digitalWrite(LED_BUILTIN, LOW);
						uint8_t localeebyte = EEPROM.readByte(i+((eeprombank-1)*256));
						if (localeebyte != metereebyte){
							writeByte(I2C_ADDR, i, localeebyte);
							updatecount++;
						}
						if (i == 255) break;
					}
					Serial.println(F(""));
					Serial.print(updatecount);
					Serial.println(F(" Bytes Written to Meters EEPROM!"));
				}
				else{
					Serial.print(F("Error Writing to I2C EEPROM @ 0x"));
					Serial.println(I2C_ADDR, HEX);
				}

			}
			else{
				Serial.println(F("\n\rRestore canceled!"));
			}
			break;
			default:
			break;
		}
		
	}
	else{
			VT100helper_MoveTo(16,1);
			VT100helper_ClearCurserDown();
			Serial.println(F("\n\rInvalid Data Entry!"));
	}
	static uint8_t updatestatus = 0;
	
}

void VT100helper_MoveTo(uint8_t Line, uint8_t Column){
	Serial.print(F("\x1b[")); //Escape
	Serial.print(Line,DEC);  //Move to Line x
	Serial.print(F(";"));
	Serial.print(Column,DEC); //Move to Column Y
	Serial.print(F("H")); 
}
void VT100helper_MoveHome(){
	Serial.print(F("\x1b[H")); //Escape & Home
}
void VT100helper_ClearCurserDown(){
	Serial.print(F("\x1b[J")); //Escape & Clear screen from cursor down
}

void VT100helper_ClearCurserLine(){
	Serial.print(F("\x1b[K")); //Escape & Clear line from Cursor to Right
}


