#include <Wire.h>
#include <util/atomic.h>
#include <avr/interrupt.h>
#include <EEPROMex.h>

const int I2C_ADDR = 0x50;

typedef enum {P_ADDRESS, P_DATA, P_CONFIRMWRITE, P_SETBANK, P_CONFIRMRESTORE} configparams;

volatile uint8_t serialstate = 0;
#define SERIALDATAENTRY 0
#define SERIALCONFIG 1
#define SERIALNEWDATAENTRYVALID 2
#define SERIALNEWADDRESSENTRYVALID 3
#define SERIALDATAINVALID 5
#define SERIALDATAREADY 6
#define SERIALESCAPE 7
#define SERIALBUTTONDEBUG    //Print verbouse iformation for undefined buttons

//Various defines
#define RETURN_NEWLINE "\n\r"
#define DATABUFFERLEN 25

uint8_t neweepromdata;
uint8_t neweepromaddress;

uint8_t eeprombank;

void MODeeprom() {

	Serial.print("\n...Flashing EEPROM...\n");
	
	//Dotless mode calibration data, copied from 0x50, 0x51
	writeByte(I2C_ADDR, 0x56, (byte) 0x83);
	writeByte(I2C_ADDR, 0x57, (byte) 0x75);

	writeByte(I2C_ADDR, 0xFB, (byte) 0x1E); //Power off after 30min
	writeByte(I2C_ADDR, 0xFC, (byte) 0xB4); //Backlight time 3min
	
	
	//Count 6200
	writeByte(I2C_ADDR, 0x12, (byte) 0x38);
	writeByte(I2C_ADDR, 0x13, (byte) 0x18);

	//Alarms disabled
	writeByte(I2C_ADDR, 0x16, (byte) 0xFF);
	writeByte(I2C_ADDR, 0x17, (byte) 0xFF);
	writeByte(I2C_ADDR, 0x18, (byte) 0xFF);
	writeByte(I2C_ADDR, 0x19, (byte) 0xFF);
	writeByte(I2C_ADDR, 0x1C, (byte) 0xFF);
	

	//With selector in 2A order:
	writeByte(I2C_ADDR, 0x87, (byte) 0x1C);  //Dotless DCA
	writeByte(I2C_ADDR, 0x97, (byte) 0x1D);  //Dotless ACA
	writeByte(I2C_ADDR, 0xA7, (byte) 0x16);  //Dot DCA
	writeByte(I2C_ADDR, 0xB7, (byte) 0x17);  //Dot ACA

	//With selector in 20A order:
	writeByte(I2C_ADDR, 0x8B, (byte) 0x18);  //Dot DCA
	writeByte(I2C_ADDR, 0x9B, (byte) 0x19);  //Dot ACA

	//With selector in NCV Display in mV
	writeByte(I2C_ADDR, 0x9C, (byte) 0x02);
	writeByte(I2C_ADDR, 0xAC, (byte) 0x1D);

	//With selector in 100A order is: (DC)A - (AC)A
	writeByte(I2C_ADDR, 0x8D, (byte) 0x1A);
	writeByte(I2C_ADDR, 0x9D, (byte) 0x1B);
	
	//With selector in V range order is: V(DC) - V(AC) - V(DC mV) - V(AC mV)
	writeByte(I2C_ADDR, 0x8E, (byte) 0x03);
	writeByte(I2C_ADDR, 0x9E, (byte) 0x04);
	writeByte(I2C_ADDR, 0xAE, (byte) 0x05);
	writeByte(I2C_ADDR, 0xBE, (byte) 0x06);
}

void printHex(int num, int precision) {
	char tmp[16];
	char format[128];

	sprintf(format, "%%.%dX", precision);

	sprintf(tmp, format, num);
	Serial.print(tmp);
}

void setup() {
	
	EEPROM.setMaxAllowedWrites(10000);
	Serial.begin(115200);
	Wire.begin();
	pinMode(LED_BUILTIN, OUTPUT);
	//Serial.print("\n...Before EEPROM Dump...\n");
	//dumpEEPROM();
	//delay(10);
	
	//MODeeprom();
	
	//delay(10);
	//Serial.print("\n...After EEPROM Dump...\n");
	//dumpEEPROM();
	printHelp();
}



void loop() {
	
	process_serial();
	
}

void dumpEEPROM()
{
	uint8_t error;
	Wire.beginTransmission(I2C_ADDR);
	error = Wire.endTransmission();
	if (error == 0){
		for (int i = 0; i < 256; i++) {
			byte b = readByte(I2C_ADDR, i);
			//Serial.print(b, HEX); Serial.print(" ");
			printHex(b, 2); Serial.print(" ");
			if ((i + 1) % 16 == 0) Serial.println();
		}
		Serial.println();
	}
	else {
		Serial.print(F("Error Reading from I2C EEPROM @ 0x"));
		Serial.println(I2C_ADDR, HEX);
	}
}

void writeByte(int i2cAddr, unsigned int addr, byte data) {
	Wire.beginTransmission(i2cAddr);
	Wire.write(addr);
	Wire.write(data);
	Wire.endTransmission();
	delay(5);
}

byte readByte(int i2cAddr, unsigned int addr) {
	byte data = 0x00;
	Wire.beginTransmission(i2cAddr);
	Wire.write(addr);
	Wire.endTransmission();
	Wire.requestFrom(i2cAddr, 1);
	while (!Wire.available()) ;
	data = Wire.read();
	return data;
}

void scanI2C(){
	uint8_t error, address;
	uint8_t nDevices;
	Serial.println("Scanning...");

	nDevices = 0;
	for(address = 1; address < 127; address++ ){
		// The i2c_scanner uses the return value of
		// the Write.endTransmisstion to see if
		// a device did acknowledge to the address.
		Wire.beginTransmission(address);
		error = Wire.endTransmission();
		if (error == 0)
		{
			Serial.print("I2C device found at address 0x");
			if (address<16)
			Serial.print("0");
			Serial.print(address,HEX);
			Serial.println("  !");
			
			nDevices++;
		}
		else if (error==4)
		{
			Serial.print("Unknown error at address 0x");
			if (address<16)
			Serial.print("0");
			Serial.println(address,HEX);
		}
	}
	if (nDevices == 0){
		Serial.println("No I2C devices found\n");
	}
	else{
		Serial.println("done\n");
	}
}

